@@ -1,5 +1,5 @@
 /******************************************************************************
-** Copyright (c) 2015-2017, Intel Corporation                                **
+** Copyright (c) 2015-2018, Intel Corporation                                **
 ** All rights reserved.                                                      **
 **                                                                           **
 ** Redistribution and use in source and binary forms, with or without        **
@@ -26,7 +26,7 @@
 ** NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS        **
 ** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              **
 ******************************************************************************/
-/* Alexander Heinecke (Intel Corp.)
+/* Alexander Heinecke, Greg Henry (Intel Corp.)
 ******************************************************************************/
 
 #ifndef GENERATOR_X86_INSTRUCTIONS_H
@@ -42,7 +42,7 @@
  * @param i_arch architecture code was generated for (needed to build clobber)
  * @param i_prefetch prefetch mode which may result in additional gp reg inits
  */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_open_stream( libxsmm_generated_code*       io_generated_code,
                                           const libxsmm_gp_reg_mapping* i_gp_reg_mapping,
                                           const char*                   i_arch,
@@ -56,12 +56,28 @@ void libxsmm_x86_instruction_open_stream( libxsmm_generated_code*       io_gener
  * @param i_arch architecture code was generated for (needed to build clobber)
  * @param i_prefetch prefetch mode which may result in additional gp reg clobbers
  */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_close_stream( libxsmm_generated_code*       io_generated_code,
                                            const libxsmm_gp_reg_mapping* i_gp_reg_mapping,
                                            const char*                   i_arch,
                                            unsigned int                  i_prefetch );
 
+/**
+ * Generates vmaskmovps/vmaskmovpd with displacements for loads and stores.
+ * Only works with i_vector_name='Y'
+ */
+LIBXSMM_API_INTERN
+void libxsmm_x86_instruction_vec_mask_move( libxsmm_generated_code* io_generated_code,
+                                     const unsigned int      i_vmove_instr,
+                                     const unsigned int      i_gp_reg_base,
+                                     const unsigned int      i_gp_reg_idx,
+                                     const unsigned int      i_scale,
+                                     const int               i_displacement,
+                                     const char              i_vector_name,
+                                     const unsigned int      i_vec_reg_number_0,
+                                     const unsigned int      i_vec_reg_mask_0,
+                                     const unsigned int      i_is_store );
+
 /**
  * Generates vmovapd/vmovupd/vmovaps/vmovups/vmovsd/vmovss/vbroadcastsd/vbroastcastss/vmovddup instructions with displacements, explicit SIB addressing is not
  * supported by this function
@@ -76,7 +92,7 @@ void libxsmm_x86_instruction_close_stream( libxsmm_generated_code*       io_gene
  * @param i_mask_reg_number the mask register to be used
  * @param i_is_store 0: load semantic, other: store semantic
  */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_vec_move( libxsmm_generated_code* io_generated_code,
                                        const unsigned int      i_instruction_set,
                                        const unsigned int      i_vmove_instr,
@@ -100,7 +116,7 @@ void libxsmm_x86_instruction_vec_move( libxsmm_generated_code* io_generated_code
  * @param i_vec_reg_number_1 the second vector register number (xmm/ymm: 0-15, zmm: 0-31)
  * @param i_vec_reg_number_2 the third vector register number (xmm/ymm: 0-15, zmm: 0-31), if this operand equals LIBXSMM_X86_VEC_REG_UNDEF -> SSE3 code generation
  */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_vec_compute_reg( libxsmm_generated_code* io_generated_code,
                                               const unsigned int      i_instruction_set,
                                               const unsigned int      i_vec_instr,
@@ -109,6 +125,27 @@ void libxsmm_x86_instruction_vec_compute_reg( libxsmm_generated_code* io_generat
                                               const unsigned int      i_vec_reg_number_1,
                                               const unsigned int      i_vec_reg_number_2 );
 
+
+/**
+ * Generates (v)XYZpd/(v)XYZps/(v)XYZsd/(v)XYZss convert instructions with 2 vector registers, memory operands are not supported as first operand
+ *
+ * @param io_generated_code pointer to the pointer of the generated code structure
+ * @param i_instruction_set requested instruction set to encode
+ * @param i_vec_instr actual operation variant
+ * @param i_vector_name the vector register name prefix (x,y or z)
+ * @param i_vec_reg_src the first/source vector register number (xmm/ymm: 0-15, zmm: 0-31)
+ * @param i_vec_reg_dst the second/destination vector register number (xmm/ymm: 0-15, zmm: 0-31)
+ * @param i_shuffle_operand is an immediate (only looked at when needed)
+ */
+LIBXSMM_API_INTERN
+void libxsmm_x86_instruction_vec_compute_convert ( libxsmm_generated_code* io_generated_code,
+                                              const unsigned int      i_instruction_set,
+                                              const unsigned int      i_vec_instr,
+                                              const char              i_vector_name,
+                                              const unsigned int      i_vec_reg_src,
+                                              const unsigned int      i_vec_reg_dst,
+                                              const unsigned int      i_shuffle_operand );
+
 /**
  * Generates (v)XYZpd/(v)XYZps/(v)XYZsd/(v)XYZss instructions with 2 vector registers, memory operand as first operand
  *
@@ -124,7 +161,26 @@ void libxsmm_x86_instruction_vec_compute_reg( libxsmm_generated_code* io_generat
  * @param i_vec_reg_number_0 the first vector register number (zmm: 0-31)
  * @param i_vec_reg_number_1 the second vector register number (zmm: 0-31)
  */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
+void libxsmm_x86_instruction_vec_compute_reg_mask( libxsmm_generated_code* io_generated_code,
+                                              const unsigned int      i_instruction_set,
+                                              const unsigned int      i_vec_instr,
+                                              const char              i_vector_name,
+                                              const unsigned int      i_vec_reg_number_0,
+                                              const unsigned int      i_vec_reg_number_1,
+                                              const unsigned int      i_vec_reg_number_2,
+                                              const unsigned int      i_mask_reg_number );
+
+/**
+ * @param i_instruction_set requested instruction set to encode
+ * @param i_vec_instr actual operation variant
+ * @param i_vector_name the vector register name prefix (z)
+ * @param i_vec_reg_number_0 the first vector register number (zmm: 0-31)
+ * @param i_vec_reg_number_1 the second vector register number (zmm: 0-31)
+ * @param i_vec_reg_number_2 the third vector register number (zmm: 0-31)
+ * @param i_mask_reg_number the mask register (0-7)
+ */
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_vec_compute_mem( libxsmm_generated_code* io_generated_code,
                                               const unsigned int      i_instruction_set,
                                               const unsigned int      i_vec_instr,
@@ -137,6 +193,61 @@ void libxsmm_x86_instruction_vec_compute_mem( libxsmm_generated_code* io_generat
                                               const unsigned int      i_vec_reg_number_0,
                                               const unsigned int      i_vec_reg_number_1 );
 
+ /**
+  * Generates quadmadd instructions added in Knights Mill
+  *
+  * @param io_generated_code pointer to the pointer of the generated code structure
+  * @param i_instruction_set requested instruction set to encode
+  * @param i_vec_instr actual operation variant
+  * @param i_gp_reg_base base address register for memory broadcast
+  * @param i_gp_reg_idx index register for memory broadcast, can be LIBXSMM_X86_GP_REG_UNDEF -> then regular displacement version is generated
+  * @param i_scale scale of index register, ignored if i_gp_reg_idx is LIBXSMM_X86_GP_REG_UNDEF
+  * @param i_displacement displacement to SIB address
+  * @param i_vector_name the vector register name prefix (z)
+  * @param i_vec_reg_number_src the second vector register number (zmm: 0-31), this define a implicit regsiter range
+  * @param i_vec_reg_number_dest the first vector register number (zmm: 0-31)
+  */
+LIBXSMM_API_INTERN
+void libxsmm_x86_instruction_vec_compute_mem_mask( libxsmm_generated_code* io_generated_code,
+                                              const unsigned int      i_instruction_set,
+                                              const unsigned int      i_vec_instr,
+                                              const unsigned int      i_use_broadcast,
+                                              const unsigned int      i_gp_reg_base,
+                                              const unsigned int      i_gp_reg_idx,
+                                              const unsigned int      i_scale,
+                                              const int               i_displacement,
+                                              const char              i_vector_name,
+                                              const unsigned int      i_vec_reg_number_0,
+                                              const unsigned int      i_vec_reg_number_1,
+                                              const unsigned int      i_shuffle_operand,
+                                              const unsigned int      i_mask_reg_number );
+
+ /**
+  * Generates quadmadd instructions added in Knights Mill
+  *
+  * @param io_generated_code pointer to the pointer of the generated code structure
+  * @param i_instruction_set requested instruction set to encode
+  * @param i_vec_instr actual operation variant
+  * @param i_gp_reg_base base address register for memory broadcast
+  * @param i_gp_reg_idx index register for memory broadcast, can be LIBXSMM_X86_GP_REG_UNDEF -> then regular displacement version is generated
+  * @param i_scale scale of index register, ignored if i_gp_reg_idx is LIBXSMM_X86_GP_REG_UNDEF
+  * @param i_displacement displacement to SIB address
+  * @param i_vector_name the vector register name prefix (z)
+  * @param i_vec_reg_number_src the second vector register number (zmm: 0-31), this define a implicit regsiter range
+  * @param i_vec_reg_number_dest the first vector register number (zmm: 0-31)
+  */
+LIBXSMM_API_INTERN
+void libxsmm_x86_instruction_vec_compute_qfma( libxsmm_generated_code* io_generated_code,
+                                               const unsigned int      i_instruction_set,
+                                               const unsigned int      i_vec_instr,
+                                               const unsigned int      i_gp_reg_base,
+                                               const unsigned int      i_gp_reg_idx,
+                                               const unsigned int      i_scale,
+                                               const int               i_displacement,
+                                               const char              i_vector_name,
+                                               const unsigned int      i_vec_reg_number_src,
+                                               const unsigned int      i_vec_reg_number_dest );
+
 /**
  * Generates shuffle instructions with 2 or 3 vector registers, memory operands are not supported as first operand
  *
@@ -148,7 +259,7 @@ void libxsmm_x86_instruction_vec_compute_mem( libxsmm_generated_code* io_generat
  * @param i_vec_reg_number_1 the second vector register number (xmm/ymm: 0-15, zmm: 0-31)
  * @param i_vec_reg_number_2 the third vector register number (xmm/ymm: 0-15, zmm: 0-31), if this operand equals LIBXSMM_X86_VEC_REG_UNDEF -> SSE3 code generation
  */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_vec_shuffle_reg( libxsmm_generated_code* io_generated_code,
                                               const unsigned int      i_instruction_set,
                                               const unsigned int      i_vec_instr,
@@ -173,7 +284,7 @@ void libxsmm_x86_instruction_vec_shuffle_reg( libxsmm_generated_code* io_generat
  * @param i_mask_reg_number the mask register (xmm/ymm: 0-15 when using AVX2), (k1-k7 when using AVX512)
  * @param i_is_gather "true" generate a gather instruction, "false" generator a scatter instruction
  */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_vec_move_gathscat( libxsmm_generated_code* io_generated_code,
                                                 const unsigned int      i_instruction_set,
                                                 const unsigned int      i_vmove_instr,
@@ -196,7 +307,7 @@ void libxsmm_x86_instruction_vec_move_gathscat( libxsmm_generated_code* io_gener
  * @param i_gp_reg_number the register number (rax=0,rcx=1,rdx=2,rbx=3,rsp=4,rbp=5,rsi=6,rdi=7,r8=8,r9=9,r10=10,r11=11,r12=12,r13=13,r14=14,r15=15) of the base address register
  * @param i_displacement the offset to the base address
  */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_prefetch( libxsmm_generated_code* io_generated_code,
                                        const unsigned int      i_prefetch_instr,
                                        const unsigned int      i_gp_reg_base,
@@ -204,6 +315,21 @@ void libxsmm_x86_instruction_prefetch( libxsmm_generated_code* io_generated_code
                                        const unsigned int      i_scale,
                                        const int               i_displacement );
 
+/**
+ * Generates alu memory movements like movq 7(%rax,%rbx,2), %rcx
+ * Takes 3 gp_registers (0-15 values)
+ * i_is_store tells whether this is a store or load
+ */
+LIBXSMM_API_INTERN
+void libxsmm_x86_instruction_alu_mem( libxsmm_generated_code* io_generated_code,
+                                      const unsigned int     i_alu_instr,
+                                      const unsigned int     i_gp_reg_base,
+                                      const unsigned int     i_gp_reg_idx,
+                                      const unsigned int     i_scale,
+                                      const int              i_displacement,
+                                      const unsigned int     i_gp_reg_number,
+                                      const unsigned int     i_is_store );
+
 /**
  * Generates regular all instructions with immediates
  *
@@ -212,12 +338,26 @@ void libxsmm_x86_instruction_prefetch( libxsmm_generated_code* io_generated_code
  * @param i_gp_reg_number the register number (rax=0,rcx=1,rdx=2,rbx=3,rsp=4,rbp=5,rsi=6,rdi=7,r8=8,r9=9,r10=10,r11=11,r12=12,r13=13,r14=14,r15=15) of the base address register
  * @param i_immediate the immediate operand
  */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_alu_imm( libxsmm_generated_code* io_generated_code,
                                       const unsigned int      i_alu_instr,
                                       const unsigned int      i_gp_reg_number,
                                       const int               i_immediate );
 
+/**
+ * Generates regular all instructions with immediates, 64bit
+ *
+ * @param io_generated_code pointer to the pointer of the generated code structure
+ * @param i_alu_instr actual alu gpr instruction
+ * @param i_gp_reg_number the register number (rax=0,rcx=1,rdx=2,rbx=3,rsp=4,rbp=5,rsi=6,rdi=7,r8=8,r9=9,r10=10,r11=11,r12=12,r13=13,r14=14,r15=15) of the base address register
+ * @param i_immediate the immediate operand
+ */
+LIBXSMM_API_INTERN
+void libxsmm_x86_instruction_alu_imm_i64( libxsmm_generated_code* io_generated_code,
+                                          const unsigned int      i_alu_instr,
+                                          const unsigned int      i_gp_reg_number,
+                                          const size_t            i_immediate );
+
 /**
  * Generates regular all instructions with immediates
  *
@@ -226,7 +366,7 @@ void libxsmm_x86_instruction_alu_imm( libxsmm_generated_code* io_generated_code,
  * @param i_gp_reg_number_src the source register number (rax=0,rcx=1,rdx=2,rbx=3,rsp=4,rbp=5,rsi=6,rdi=7,r8=8,r9=9,r10=10,r11=11,r12=12,r13=13,r14=14,r15=15) of the base address register
  * @param i_gp_reg_number_dest the destination register number (rax=0,rcx=1,rdx=2,rbx=3,rsp=4,rbp=5,rsi=6,rdi=7,r8=8,r9=9,r10=10,r11=11,r12=12,r13=13,r14=14,r15=15) of the base address register
  */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_alu_reg( libxsmm_generated_code* io_generated_code,
                                       const unsigned int      i_alu_instr,
                                       const unsigned int      i_gp_reg_number_src,
@@ -238,7 +378,7 @@ void libxsmm_x86_instruction_alu_reg( libxsmm_generated_code* io_generated_code,
  * @param io_generated_code pointer to the pointer of the generated code structure
  * @param i_gp_reg_number the source register number (rax=0,rcx=1,rdx=2,rbx=3,rsp=4,rbp=5,rsi=6,rdi=7,r8=8,r9=9,r10=10,r11=11,r12=12,r13=13,r14=14,r15=15)
  */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_push_reg( libxsmm_generated_code* io_generated_code,
                                        const unsigned int      i_gp_reg_number );
 
@@ -248,7 +388,7 @@ void libxsmm_x86_instruction_push_reg( libxsmm_generated_code* io_generated_code
  * @param io_generated_code pointer to the pointer of the generated code structure
  * @param i_gp_reg_number the source register number (rax=0,rcx=1,rdx=2,rbx=3,rsp=4,rbp=5,rsi=6,rdi=7,r8=8,r9=9,r10=10,r11=11,r12=12,r13=13,r14=14,r15=15)
  */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_pop_reg( libxsmm_generated_code* io_generated_code,
                                       const unsigned int      i_gp_reg_number );
 
@@ -260,7 +400,7 @@ void libxsmm_x86_instruction_pop_reg( libxsmm_generated_code* io_generated_code,
  * @param i_gp_reg_number the register number (rax=0,rcx=1,rdx=2,rbx=3,rsp=4,rbp=5,rsi=6,rdi=7,r8=8,r9=9,r10=10,r11=11,r12=12,r13=13,r14=14,r15=15) of the base address register
  * @param i_mask_reg_number the register number (k1=1...k7=7)
  */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_mask_move( libxsmm_generated_code* io_generated_code,
                                         const unsigned int      i_mask_instr,
                                         const unsigned int      i_gp_reg_number,
@@ -275,7 +415,7 @@ void libxsmm_x86_instruction_mask_move( libxsmm_generated_code* io_generated_cod
  * @param i_mask_reg_number_src_1 the second operand register number (att syntax) (k1=1...k7=7)
  * @param i_mask_reg_number_dest the third operand register number (att syntax) (k1=1...k7=7)
  */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_mask_compute_reg( libxsmm_generated_code* io_generated_code,
                                                const unsigned int      i_mask_instr,
                                                const unsigned int      i_mask_reg_number_src_0,
@@ -288,7 +428,7 @@ void libxsmm_x86_instruction_mask_compute_reg( libxsmm_generated_code* io_genera
  * @param io_generated_code pointer to the pointer of the generated code structure
  * @param io_loop_label_tracker data structure to handle loop labels, nested loops are supported, but not overlapping loops
 */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_register_jump_label( libxsmm_generated_code*     io_generated_code,
                                                   libxsmm_loop_label_tracker* io_loop_label_tracker );
 
@@ -299,7 +439,7 @@ void libxsmm_x86_instruction_register_jump_label( libxsmm_generated_code*     io
  * @param i_jmp_instr the particular jump instruction used
  * @param io_loop_label_tracker data structure to handle loop labels will jump to latest registered label
 */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_jump_back_to_label( libxsmm_generated_code*     io_generated_code,
                                                  const unsigned int          i_jmp_instr,
                                                  libxsmm_loop_label_tracker* io_loop_label_tracker );
@@ -314,7 +454,7 @@ void libxsmm_x86_instruction_jump_back_to_label( libxsmm_generated_code*     io_
  * @param i_vector_name the vector register name prefix (x,y or z)
  * @param i_vec_reg_number the destination(gather)/source(scatter) vec register (xmm/ymm: 0-15, zmm: 0-31)
 */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_full_vec_load_of_constants ( libxsmm_generated_code *io_generated_code,
                                                           const unsigned char *i_data,
                                                           const char *i_id,
@@ -327,7 +467,7 @@ void libxsmm_x86_instruction_full_vec_load_of_constants ( libxsmm_generated_code
  * @param io_generated_code pointer to the pointer of the generated code structure
  * @param i_arch architecture code was generated for (needed to build clobber)
  */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_open_stream_convolution( libxsmm_generated_code*                   io_generated_code,
                                                       const unsigned int                        i_gp_reg_input,
                                                       const unsigned int                        i_gp_reg_weight,
@@ -344,9 +484,64 @@ void libxsmm_x86_instruction_open_stream_convolution( libxsmm_generated_code*
  * @param io_generated_code pointer to the pointer of the generated code structure
  * @param i_arch architecture code was generated for (needed to build clobber)
  */
-LIBXSMM_INTERNAL_API
+LIBXSMM_API_INTERN
 void libxsmm_x86_instruction_close_stream_convolution( libxsmm_generated_code*       io_generated_code,
                                                 const char*                   i_arch );
 
+
+/**
+ * @TODO: clean-up
+ * Opens the inline assembly section / jit stream for matcopy, this is hacked and should be cleaned up
+ *
+ * @param io_generated_code pointer to the pointer of the generated code structure
+ * @param i_arch architecture code was generated for (needed to build clobber)
+ */
+LIBXSMM_API_INTERN
+void libxsmm_x86_instruction_open_stream_matcopy( libxsmm_generated_code*                   io_generated_code,
+                                                  const unsigned int                        i_gp_reg_a,
+                                                  const unsigned int                        i_gp_reg_lda,
+                                                  const unsigned int                        i_gp_reg_b,
+                                                  const unsigned int                        i_gp_reg_ldb,
+                                                  const unsigned int                        i_gp_reg_a_pf,
+                                                  const unsigned int                        i_gp_reg_b_pf,
+                                                  const char*                               i_arch );
+
+/**
+ * @TODO: clean-up
+ * Closes the inline assembly section / jit stream for matcopy, this is hacked and should be cleaned up
+ *
+ * @param io_generated_code pointer to the pointer of the generated code structure
+ * @param i_arch architecture code was generated for (needed to build clobber)
+ */
+LIBXSMM_API_INTERN
+void libxsmm_x86_instruction_close_stream_matcopy( libxsmm_generated_code*       io_generated_code,
+                                                   const char*                   i_arch );
+
+/**
+ * @TODO: clean-up
+ * Opens the inline assembly section / jit stream for transposes, this is hacked and should be cleaned up
+ *
+ * @param io_generated_code pointer to the pointer of the generated code structure
+ * @param i_arch architecture code was generated for (needed to build clobber)
+ */
+LIBXSMM_API_INTERN
+void libxsmm_x86_instruction_open_stream_transpose( libxsmm_generated_code*                   io_generated_code,
+                                                    const unsigned int                        i_gp_reg_a,
+                                                    const unsigned int                        i_gp_reg_lda,
+                                                    const unsigned int                        i_gp_reg_b,
+                                                    const unsigned int                        i_gp_reg_ldb,
+                                                    const char*                               i_arch );
+
+/**
+ * @TODO: clean-up
+ * Closes the inline assembly section / jit stream for transposes, this is hacked and should be cleaned up
+ *
+ * @param io_generated_code pointer to the pointer of the generated code structure
+ * @param i_arch architecture code was generated for (needed to build clobber)
+ */
+LIBXSMM_API_INTERN
+void libxsmm_x86_instruction_close_stream_transpose( libxsmm_generated_code*       io_generated_code,
+                                                     const char*                   i_arch );
+
 #endif /* GENERATOR_X86_INSTRUCTIONS_H */
 
