@@ -26,12 +26,21 @@
 ** NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS        **
 ** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              **
 ******************************************************************************/
-/* Rajkishore Barik (Intel Corp.), Alexander Heinecke (Intel Corp.)
+/* Rajkishore Barik, Ankush Mandal, Alexander Heinecke (Intel Corp.)
 ******************************************************************************/
 #include "libxsmm_dnn_convolution_backward.h"
-#include <libxsmm.h>
+#include <libxsmm_intrinsics_x86.h>
+#include "libxsmm_main.h"
 
-LIBXSMM_API_DEFINITION libxsmm_dnn_err_t libxsmm_dnn_convolve_st_bwd_custom_custom(libxsmm_dnn_layer* handle, int start_thread, int tid)
+#if defined(LIBXSMM_OFFLOAD_TARGET)
+# pragma offload_attribute(push,target(LIBXSMM_OFFLOAD_TARGET))
+#endif
+#include <string.h>
+#if defined(LIBXSMM_OFFLOAD_TARGET)
+# pragma offload_attribute(pop)
+#endif
+
+LIBXSMM_API_INTERN libxsmm_dnn_err_t libxsmm_dnn_convolve_st_bwd_custom_custom(libxsmm_dnn_layer* handle, int start_thread, int tid)
 {
   libxsmm_dnn_err_t status = LIBXSMM_DNN_SUCCESS;
 
@@ -41,13 +50,58 @@ LIBXSMM_API_DEFINITION libxsmm_dnn_err_t libxsmm_dnn_convolve_st_bwd_custom_cust
     return status;
   }
 
+  /* for low/mixed precision we need some scratch to be bound */
+  if ( (handle->datatype != handle->datatype_itm) && (handle->scratch7 == 0) ) {
+    status = LIBXSMM_DNN_ERR_DATA_NOT_BOUND;
+    return status;
+  }
+
   /* check if we have a kernel JITed */
   if (handle->code_bwd[0].xconv.sconv == 0) {
     if (handle->datatype == LIBXSMM_DNN_DATATYPE_F32 && handle->datatype_itm == LIBXSMM_DNN_DATATYPE_F32 ) {
       typedef float element_input_type;
       typedef float element_output_type;
       typedef float element_filter_type;
-# include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_fallback.tpl.c"
+      if (handle->padding_flag == 1) {
+#define INPUT_PADDING
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_fallback.tpl.c"
+#undef INPUT_PADDING
+      } else {
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_fallback.tpl.c"
+      }
+    } else if (handle->datatype ==  LIBXSMM_DNN_DATATYPE_I16 && handle->datatype_itm == LIBXSMM_DNN_DATATYPE_I32 ) {
+      typedef int element_input_type;
+      typedef short element_output_type;
+      typedef short element_filter_type;
+      if (handle->padding_flag == 1) {
+#define INPUT_PADDING
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_fallback.tpl.c"
+#undef INPUT_PADDING
+      } else {
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_fallback.tpl.c"
+      }
+    } else if (handle->datatype == LIBXSMM_DNN_DATATYPE_I8 && handle->datatype_itm == LIBXSMM_DNN_DATATYPE_I16 && (handle->desc.options & LIBXSMM_DNN_CONV_OPTION_ACTIVATION_UNSIGNED) > 0 ) {
+      typedef unsigned short element_input_type;
+      typedef char element_output_type;
+      typedef char element_filter_type;
+      if (handle->padding_flag == 1) {
+#define INPUT_PADDING
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_fallback.tpl.c"
+#undef INPUT_PADDING
+      } else {
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_fallback.tpl.c"
+      }
+    } else if (handle->datatype == LIBXSMM_DNN_DATATYPE_I8 && handle->datatype_itm == LIBXSMM_DNN_DATATYPE_I32 && (handle->desc.options & LIBXSMM_DNN_CONV_OPTION_ACTIVATION_UNSIGNED) > 0 ) {
+      typedef unsigned int element_input_type;
+      typedef char element_output_type;
+      typedef char element_filter_type;
+      if (handle->padding_flag == 1) {
+#define INPUT_PADDING
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_fallback.tpl.c"
+#undef INPUT_PADDING
+      } else {
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_fallback.tpl.c"
+      }
     } else {
       status = LIBXSMM_DNN_ERR_UNSUPPORTED_DATATYPE;
       return status;
@@ -58,11 +112,17 @@ LIBXSMM_API_DEFINITION libxsmm_dnn_err_t libxsmm_dnn_convolve_st_bwd_custom_cust
 #if 0
       if (handle->desc.N*handle->blocksifm >= handle->desc.threads) {
 #endif
-        typedef float element_input_type;
-        typedef float element_output_type;
-        typedef float element_filter_type;
-        typedef libxsmm_sconvfunction libxsmm_convfunction;
-# include "template/libxsmm_dnn_convolve_st_bwd_custom_custom.tpl.c"
+      typedef float element_input_type;
+      typedef float element_output_type;
+      typedef float element_filter_type;
+      typedef libxsmm_sconvfunction libxsmm_convfunction;
+      if (handle->padding_flag == 1) {
+#define INPUT_PADDING
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom.tpl.c"
+#undef INPUT_PADDING
+      } else {
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom.tpl.c"
+      }
 #if 0
       }
       else {
@@ -70,9 +130,45 @@ LIBXSMM_API_DEFINITION libxsmm_dnn_err_t libxsmm_dnn_convolve_st_bwd_custom_cust
         typedef float element_output_type;
         typedef float element_filter_type;
         typedef libxsmm_sconvfunction libxsmm_convfunction;
-# include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_img_par.tpl.c"
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_img_par.tpl.c"
       }
 #endif
+    } else if (handle->datatype ==  LIBXSMM_DNN_DATATYPE_I16 && handle->datatype_itm == LIBXSMM_DNN_DATATYPE_I32 ) {
+      typedef int element_input_type;
+      typedef short element_output_type;
+      typedef short element_filter_type;
+      typedef libxsmm_wconvfunction_bwd libxsmm_convfunction;
+      if (handle->padding_flag == 1) {
+#define INPUT_PADDING
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_1.tpl.c"
+#undef INPUT_PADDING
+      } else {
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_1.tpl.c"
+      }
+    } else if (handle->datatype == LIBXSMM_DNN_DATATYPE_I8 && handle->datatype_itm == LIBXSMM_DNN_DATATYPE_I16 && (handle->desc.options & LIBXSMM_DNN_CONV_OPTION_ACTIVATION_UNSIGNED) > 0 ) {
+      typedef unsigned short element_input_type;
+      typedef char element_output_type;
+      typedef char element_filter_type;
+      typedef libxsmm_busconvfunction_bwd libxsmm_convfunction;
+      if (handle->padding_flag == 1) {
+#define INPUT_PADDING
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_1.tpl.c"
+#undef INPUT_PADDING
+      } else {
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_1.tpl.c"
+      }
+    } else if (handle->datatype == LIBXSMM_DNN_DATATYPE_I8 && handle->datatype_itm == LIBXSMM_DNN_DATATYPE_I32 && (handle->desc.options & LIBXSMM_DNN_CONV_OPTION_ACTIVATION_UNSIGNED) > 0 ) {
+      typedef unsigned int element_input_type;
+      typedef char element_output_type;
+      typedef char element_filter_type;
+      typedef libxsmm_budconvfunction_bwd libxsmm_convfunction;
+      if (handle->padding_flag == 1) {
+#define INPUT_PADDING
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_1.tpl.c"
+#undef INPUT_PADDING
+      } else {
+#include "template/libxsmm_dnn_convolve_st_bwd_custom_custom_1.tpl.c"
+      }
     } else {
       status = LIBXSMM_DNN_ERR_UNSUPPORTED_DATATYPE;
       return status;
@@ -82,7 +178,7 @@ LIBXSMM_API_DEFINITION libxsmm_dnn_err_t libxsmm_dnn_convolve_st_bwd_custom_cust
   return status;
 }
 
-LIBXSMM_API_DEFINITION libxsmm_dnn_err_t libxsmm_dnn_convolve_st_bwd_nhwc_rsck(libxsmm_dnn_layer* handle, int start_thread, int tid)
+LIBXSMM_API_INTERN libxsmm_dnn_err_t libxsmm_dnn_convolve_st_bwd_nhwc_rsck(libxsmm_dnn_layer* handle, int start_thread, int tid)
 {
   libxsmm_dnn_err_t status = LIBXSMM_DNN_SUCCESS;
 
@@ -98,7 +194,13 @@ LIBXSMM_API_DEFINITION libxsmm_dnn_err_t libxsmm_dnn_convolve_st_bwd_nhwc_rsck(l
       typedef float element_input_type;
       typedef float element_output_type;
       typedef float element_filter_type;
+      if (handle->padding_flag == 1) {
+#define INPUT_PADDING
+#include "template/libxsmm_dnn_convolve_st_bwd_nhwc_rsck_fallback.tpl.c"
+#undef INPUT_PADDING
+      } else {
 # include "template/libxsmm_dnn_convolve_st_bwd_nhwc_rsck_fallback.tpl.c"
+      }
     } else {
       status = LIBXSMM_DNN_ERR_UNSUPPORTED_DATATYPE;
       return status;
@@ -113,7 +215,13 @@ LIBXSMM_API_DEFINITION libxsmm_dnn_err_t libxsmm_dnn_convolve_st_bwd_nhwc_rsck(l
         typedef float element_output_type;
         typedef float element_filter_type;
         typedef libxsmm_sconvfunction libxsmm_convfunction;
+        if (handle->padding_flag == 1) {
+#define INPUT_PADDING
+#include "template/libxsmm_dnn_convolve_st_bwd_nhwc_rsck.tpl.c"
+#undef INPUT_PADDING
+        } else {
 # include "template/libxsmm_dnn_convolve_st_bwd_nhwc_rsck.tpl.c"
+        }
 #if 0
       }
       else {
@@ -133,7 +241,7 @@ LIBXSMM_API_DEFINITION libxsmm_dnn_err_t libxsmm_dnn_convolve_st_bwd_nhwc_rsck(l
   return status;
 }
 
-LIBXSMM_API_DEFINITION libxsmm_dnn_err_t libxsmm_dnn_convolve_st_bwd_nhwc_custom(libxsmm_dnn_layer* handle, int start_thread, int tid)
+LIBXSMM_API_INTERN libxsmm_dnn_err_t libxsmm_dnn_convolve_st_bwd_nhwc_custom(libxsmm_dnn_layer* handle, int start_thread, int tid)
 {
   libxsmm_dnn_err_t status = LIBXSMM_DNN_SUCCESS;
 
@@ -149,7 +257,13 @@ LIBXSMM_API_DEFINITION libxsmm_dnn_err_t libxsmm_dnn_convolve_st_bwd_nhwc_custom
       typedef float element_input_type;
       typedef float element_output_type;
       typedef float element_filter_type;
+      if (handle->padding_flag == 1) {
+#define INPUT_PADDING
 # include "template/libxsmm_dnn_convolve_st_bwd_nhwc_custom_fallback.tpl.c"
+#undef INPUT_PADDING
+      } else {
+# include "template/libxsmm_dnn_convolve_st_bwd_nhwc_custom_fallback.tpl.c"
+      }
     } else {
       status = LIBXSMM_DNN_ERR_UNSUPPORTED_DATATYPE;
       return status;
@@ -164,7 +278,13 @@ LIBXSMM_API_DEFINITION libxsmm_dnn_err_t libxsmm_dnn_convolve_st_bwd_nhwc_custom
         typedef float element_output_type;
         typedef float element_filter_type;
         typedef libxsmm_sconvfunction libxsmm_convfunction;
+        if (handle->padding_flag == 1) {
+#define INPUT_PADDING
 # include "template/libxsmm_dnn_convolve_st_bwd_nhwc_custom.tpl.c"
+#undef INPUT_PADDING
+        } else {
+# include "template/libxsmm_dnn_convolve_st_bwd_nhwc_custom.tpl.c"
+        }
 #if 0
       }
       else {
